from __future__ import annotations

import asyncio
import queue
from typing import NamedTuple, Tuple, Optional, Set
from urllib.parse import urlencode

import httpx
from rich.markdown import Markdown
from textual.app import ComposeResult
from textual.widgets import (
    Input
)

from mys_goods_tool.api import create_mobile_captcha, create_mmt, get_login_ticket_by_captcha, \
    get_multi_token_by_login_ticket, get_cookie_token_by_stoken, get_stoken_v2_by_v1, get_ltoken_by_stoken
from mys_goods_tool.custom_css import *
from mys_goods_tool.custom_widget import RadioStatus, StaticStatus, ControllableButton, LoadingDisplay
from mys_goods_tool.data_model import GeetestResult, MmtData, GetCookieStatus
from mys_goods_tool.exchange_plan_view import ExchangePlanView
from mys_goods_tool.geetest import GeetestProcessManager, SetAddressProcessManager
from mys_goods_tool.user_data import config as conf, UserAccount, CONFIG_PATH
from mys_goods_tool.utils import logger


class LoginView(Container):
    """
    登录页面
    """
    DEFAULT_CSS = """
    LoginView {
        height: auto;
    }
    """

    def compose(self) -> ComposeResult:
        yield CaptchaLoginInformation()
        yield PhoneForm()
        yield CaptchaForm()


class CaptchaLoginInformation(Container):
    """
    短信验证登录页面的信息提示区域
    """
    DEFAULT_CSS = """
    CaptchaLoginInformation {
        height: auto;
        margin: 1 0;
        overflow: hidden;
    }

    CaptchaLoginInformation Horizontal {
        align: center middle;
    }
    """

    RadioTuple = NamedTuple("RadioTuple",
                            create_geetest=RadioStatus,
                            http_server=RadioStatus,
                            geetest_finished=RadioStatus,
                            create_captcha=RadioStatus,
                            login_ticket_by_captcha=RadioStatus,
                            multi_token_by_login_ticket=RadioStatus,
                            get_stoken_v2=RadioStatus,
                            get_ltoken_by_stoken=RadioStatus,
                            cookie_token_by_stoken=RadioStatus,
                            login_finished=RadioStatus
                            )

    StaticTuple = NamedTuple("StaticTuple",
                             geetest_title=Static,
                             geetest_text=StaticStatus
                             )

    radio_tuple = RadioTuple(
        create_geetest=RadioStatus("短信验证码 - 申请人机验证任务"),
        http_server=RadioStatus("开启人机验证网页服务器"),
        geetest_finished=RadioStatus("完成人机验证"),
        create_captcha=RadioStatus("发出短信验证码"),
        login_ticket_by_captcha=RadioStatus("通过验证码获取 login_ticket"),
        multi_token_by_login_ticket=RadioStatus("通过 login_ticket 获取 stoken"),
        get_stoken_v2=RadioStatus("获取 v2 版本 stoken 和 mid"),
        get_ltoken_by_stoken=RadioStatus("通过 stoken_v2 获取 ltoken"),
        cookie_token_by_stoken=RadioStatus("通过 stoken_v2 获取 cookie_token"),
        login_finished=RadioStatus("完成登录")
    )

    SAVE_TEXT = str(CONFIG_PATH)
    GEETEST_TEXT = "- 暂无需要完成的人机验证任务 -"

    static_tuple = StaticTuple(
        geetest_title=Static(Markdown("## GEETEST人机验证链接")),
        geetest_text=StaticStatus(GEETEST_TEXT)
    )

    radio_set = CaptchaStepSet(*radio_tuple)
    static_set = CaptchaTips(*static_tuple)

    def compose(self) -> ComposeResult:
        yield Horizontal(self.radio_set, self.static_set)


class PhoneForm(LoginForm):
    """
    手机号 表单
    """
    input = Input(placeholder="手机号", id="login_phone")
    """手机号输入框"""
    device_id: Optional[str] = None
    """人机验证过程的设备ID"""
    client: Optional[httpx.AsyncClient] = None
    """人机验证过程的连接对象"""

    ButtonTuple = NamedTuple("ButtonTuple",
                             send=ControllableButton,
                             stop_geetest=ControllableButton,
                             success=ControllableButton,
                             error=ControllableButton)

    def __init__(self):
        super().__init__()

        self.mmt_data: Optional[MmtData] = None
        """GEETEST行为验证任务数据"""
        self.geetest_manager: Optional[GeetestProcessManager] = None
        """包含进程池的GEETEST验证HTTP服务器 进程管理器"""
        self.set_address_manager = SetAddressProcessManager(self.set_address_callback,
                                                            self.set_address_error_callback)
        """包含进程池的可用监听地址获取 进程管理器"""
        self.loop = asyncio.get_event_loop()
        """事件循环"""
        self.loop_tasks: Set[asyncio.Task] = set()
        """异步任务集合（保留其强引用）"""
        self.before_create_captcha = True
        """当前状态是否处于按下“发送短信验证码”按钮之前"""

        self.loading = LoadingDisplay()
        self.loading.hide()

        self.button = self.ButtonTuple(
            send=ControllableButton("发送短信验证码", variant="primary", id="create_captcha_send"),
            stop_geetest=ControllableButton("放弃人机验证", variant="warning", id="create_captcha_stop_geetest"),
            success=ControllableButton("完成", variant="success", id="create_captcha_success"),
            error=ControllableButton("返回", variant="error", id="create_captcha_error")
        )
        for button in self.button[1:]:
            button.hide()

    def compose(self) -> ComposeResult:
        yield Static("手机号", classes="label")
        yield self.input
        yield Static()
        yield from self.button
        yield Static()
        yield self.loading

    def close_create_captcha_send(self):
        """
        关闭发送短信验证码按钮
        """
        self.loading.hide()
        self.button.send.hide()
        self.button.send.enable()

    def httpd_error_callback(self, exception: BaseException):
        """
        GEETEST验证HTTP服务器启动失败时的回调函数
        """
        logger.error("用于Geetest验证的HTTP服务器启动失败")
        logger.debug(exception)
        CaptchaLoginInformation.radio_tuple.http_server.turn_off()
        self.button.stop_geetest.hide()
        self.button.error.show()

    async def listen_result(self):
        """
        等待GEETEST验证结果的异步任务
        """
        self.app.notice("请前往链接进行验证")
        while True:
            await asyncio.sleep(1)
            try:
                geetest_result: GeetestResult = self.geetest_manager.result_queue.get_nowait()
            except queue.Empty:
                continue
            else:
                logger.info(f"已收到Geetest验证结果数据，将发送验证码至 {self.input.value}")
                CaptchaLoginInformation.radio_tuple.geetest_finished.turn_on()
                self.loading.show()
                create_captcha_status, PhoneForm.client = await create_mobile_captcha(int(self.input.value),
                                                                                      self.mmt_data,
                                                                                      geetest_result,
                                                                                      PhoneForm.client,
                                                                                      device_id=PhoneForm.device_id)
                if create_captcha_status:
                    self.loading.hide()
                    logger.info(f"短信验证码已发送至 {self.input.value}")
                    CaptchaLoginInformation.radio_tuple.create_captcha.turn_on()
                    CaptchaLoginInformation.static_tuple.geetest_text.change_text(CaptchaLoginInformation.GEETEST_TEXT,
                                                                                  "center")
                    self.button.success.show()
                    self.button.stop_geetest.hide()

                    self.geetest_manager.pipe[1].send(True)
                    await self.geetest_manager.force_stop_later(10)

                    self.app.notice("短信验证码已发送至 [green]" + self.input.value + "[/]")
                    break
                else:
                    self.loading.hide()
                    self.button.error.show()
                    self.button.stop_geetest.hide()
                    CaptchaLoginInformation.static_tuple.geetest_text.change_text(CaptchaLoginInformation.GEETEST_TEXT,
                                                                                  "center")
                    self.app.notice("[red]短信验证码发送失败[/]")

    def set_address_callback(self, address: Tuple[str, int]):
        """
        可用监听地址获取成功时的回调函数

        :param address: 返回的可用地址
        """
        if not address:
            self.close_create_captcha_send()
            self.button.error.show()
            return
        self.geetest_manager = GeetestProcessManager(address, error_httpd_callback=self.httpd_error_callback)
        logger.info(f"尝试在 http://{address[0]}:{address[1]} 上启动用于Geetest验证的HTTP服务器")
        self.geetest_manager.start()

        self.close_create_captcha_send()
        self.button.stop_geetest.show()
        CaptchaLoginInformation.radio_tuple.http_server.turn_on()

        task = self.loop.create_task(self.listen_result())
        self.loop_tasks.add(task)
        task.add_done_callback(self.loop_tasks.discard)

        params = {
            "gt": self.mmt_data.gt,
            "mmtKey": self.mmt_data.mmt_key,
            "riskType": self.mmt_data.risk_type
        }
        url_params = urlencode(params)
        link = f"http://{address[0]}:{address[1]}/index.html?{url_params}"
        link_localized = f"http://{address[0]}:{address[1]}/localized.html?{url_params}"
        CaptchaLoginInformation.static_tuple.geetest_text.change_text(
            renderable=f"\n- 请前往链接进行验证：\n"
                       f"[@click=app.open_link('{link}')]{link}[/]\n"
                       f"\n- 如果页面加载慢或者出错，尝试：\n"
                       f"[@click=app.open_link('{link_localized}')]{link_localized}[/]",
            text_align="left")
        logger.info(f"请前往链接进行人机验证：{link}")
        logger.info(f"如果页面加载慢或者出错，尝试：{link_localized}")

    def set_address_error_callback(self, exception: BaseException):
        """
        可用监听地址获取失败时的回调函数
        """
        logger.error("尝试获取可用HTTP监听地址失败")
        logger.debug(exception)
        self.close_create_captcha_send()
        self.button.error.show()
        self.app.notice("[red]尝试获取可用HTTP监听地址失败！[/]")
        return

    async def create_captcha(self):
        """
        发送验证码的完整操作
        """
        if not self.before_create_captcha:
            return
        elif not self.input.value:
            self.app.notice("登录信息缺少 [bold red]手机号[/] ！")
            return
        self.before_create_captcha = False

        [i.turn_off() for i in CaptchaLoginInformation.radio_tuple]
        self.button.send.disable()
        self.loading.show()

        create_mmt_status, self.mmt_data, PhoneForm.device_id, PhoneForm.client = await create_mmt()
        if not create_mmt_status:
            self.close_create_captcha_send()
            self.button.error.show()
            self.app.notice("[red]获取Geetest行为验证任务数据失败！[/]")
            return
        else:
            logger.info(f"已成功获取Geetest行为验证任务数据 {self.mmt_data}")
            CaptchaLoginInformation.radio_tuple.create_geetest.turn_on()
            self.set_address_manager.start()

        return create_mmt_status

    async def on_input_submitted(self, _: Input.Submitted):
        await self.create_captcha()

    async def _on_button_pressed(self, event: ControllableButton.Pressed):
        if event.button.id == "create_captcha_send":
            # 按下“发送短信验证码”按钮时触发的事件

            await self.create_captcha()

        elif event.button.id == "create_captcha_stop_geetest":
            # 按下“放弃人机验证”按钮时触发的事件

            CaptchaLoginInformation.static_tuple.geetest_text.change_text(CaptchaLoginInformation.GEETEST_TEXT,
                                                                          "center")
            [i.turn_off() for i in CaptchaLoginInformation.radio_tuple]
            self.geetest_manager.pipe[1].send(True)
            self.button.stop_geetest.hide()
            self.button.send.show()
            self.before_create_captcha = True
            await self.geetest_manager.force_stop_later(10)

        elif event.button.id in ["create_captcha_success", "create_captcha_error"]:
            # 按下“完成（成功）”或“返回（出错）”按钮时触发的事件

            if event.button.id == "create_captcha_error":
                [i.turn_off() for i in CaptchaLoginInformation.radio_tuple]
            self.button.success.hide()
            self.button.error.hide()
            self.button.send.show()
            self.before_create_captcha = True


class CaptchaForm(LoginForm):
    """
    验证码 表单
    """
    ButtonTuple = NamedTuple("ButtonTuple", login=ControllableButton, success=ControllableButton,
                             error=ControllableButton)

    def __init__(self):
        super().__init__()
        self.login_result: Optional[GetCookieStatus] = None
        """登录操作返回值"""
        self.before_login: bool = True
        """当前状态是否在登录操作之前（不处于正在登录的状态）"""

        self.input = Input(placeholder="若发送验证码失败，也可前往米哈游通信证页面手动发送", id="login_captcha")

        self.loading = LoadingDisplay()
        self.loading.hide()

        self.button = self.ButtonTuple(
            login=ControllableButton("登录 / 刷新Cookies", variant="primary", id="login"),
            success=ControllableButton("完成", variant="success", id="login_success"),
            error=ControllableButton("返回", variant="error", id="login_error")
        )
        for i in self.button[1:]:
            i.hide()

    def compose(self) -> ComposeResult:
        yield Static("验证码", classes="label")
        yield self.input
        yield Static()
        yield from self.button
        yield Static()
        yield self.loading

    def close_login(self):
        self.button.login.hide()
        self.button.login.enable()

    async def login(self):
        """
        登录的完整操作
        """
        if not self.before_login:
            return
        elif not PhoneForm.input.value:
            self.app.notice("登录信息缺少 [bold red]手机号[/] ！")
            return
        elif not self.input.value.isdigit() and self.input.value:
            self.app.notice("登录信息 [bold red]验证码[/] 需要是数字或为空（刷新Cookies）！")
            return
        self.before_login = False

        self.button.login.disable()
        self.loading.show()

        account: Optional[UserAccount] = None
        login_status: GetCookieStatus = GetCookieStatus(success=False)
        phone_number = PhoneForm.input.value
        captcha = int(self.input.value) if self.input.value.isdigit() else self.input.value

        # 1. 通过短信验证码获取 login_ticket / 使用已有 login_ticket
        if captcha:
            login_status, cookies = await get_login_ticket_by_captcha(phone_number,
                                                                      captcha,
                                                                      PhoneForm.client,
                                                                      PhoneForm.device_id)
            if login_status:
                logger.info(f"用户 {phone_number} 成功获取 login_ticket: {cookies.login_ticket}")
                account = conf.accounts.get(cookies.bbs_uid)
                """当前的账户数据对象"""
                if not account or not account.cookies:
                    conf.accounts.update({
                        cookies.bbs_uid: UserAccount(phone_number=phone_number, cookies=cookies)
                    })
                    account = conf.accounts[cookies.bbs_uid]
                else:
                    account.cookies.update(cookies)
                conf.save()
                CaptchaLoginInformation.radio_tuple.login_ticket_by_captcha.turn_on()
        else:
            account_list = list(filter(lambda x: x.phone_number == phone_number, conf.accounts.values()))
            account = account_list[0] if account_list else None
            if not account:
                self.app.notice(f"手机号为 [bold red]{phone_number}[/] 的账户暂未被绑定！")
                self.loading.hide()
                self.button.error.show()
                self.close_login()
                return

        # 2. 通过 login_ticket 获取 stoken 和 ltoken
        if login_status or account:
            login_status, cookies = await get_multi_token_by_login_ticket(account.cookies)
            if login_status:
                logger.info(f"用户 {phone_number} 成功获取 stoken: {cookies.stoken}")
                account.cookies.update(cookies)
                conf.save()
                CaptchaLoginInformation.radio_tuple.multi_token_by_login_ticket.turn_on()

                # 3. 通过 stoken_v1 获取 stoken_v2 和 mid
                login_status, cookies = await get_stoken_v2_by_v1(account.cookies, account.device_id_ios)
                if login_status:
                    logger.info(f"用户 {phone_number} 成功获取 stoken_v2: {cookies.stoken_v2}")
                    account.cookies.update(cookies)
                    conf.save()
                    CaptchaLoginInformation.radio_tuple.get_stoken_v2.turn_on()

                    # 4. 通过 stoken_v2 获取 ltoken
                    login_status, cookies = await get_ltoken_by_stoken(account.cookies, account.device_id_ios)
                    if login_status:
                        logger.info(f"用户 {phone_number} 成功获取 ltoken: {cookies.ltoken}")
                        account.cookies.update(cookies)
                        conf.save()
                        CaptchaLoginInformation.radio_tuple.get_ltoken_by_stoken.turn_on()

                        # 5. 通过 stoken_v2 获取 cookie_token
                        login_status, cookies = await get_cookie_token_by_stoken(account.cookies, account.device_id_ios)
                        if login_status:
                            logger.info(f"用户 {phone_number} 成功获取 cookie_token: {cookies.cookie_token}")
                            account.cookies.update(cookies)
                            conf.save()
                            CaptchaLoginInformation.radio_tuple.cookie_token_by_stoken.turn_on()

                            # TODO 2023/04/12 此处如果可以模拟App的登录操作，再标记为登录完成，更安全
                            CaptchaLoginInformation.radio_tuple.login_finished.turn_on()
                            self.app.notice(f"用户 [bold green]{phone_number}[/] 登录成功！")
                            self.button.success.show()

                            ExchangePlanView.account_content.update_data()

        self.loading.hide()
        if not login_status:
            notice_text = "登录失败：[bold red]"
            if login_status.incorrect_captcha:
                notice_text += "验证码错误！"
            elif login_status.login_expired:
                notice_text += "登录失效！"
            elif login_status.incorrect_return:
                notice_text += "服务器返回错误！"
            elif login_status.network_error:
                notice_text += "网络连接失败！"
            elif login_status.missing_bbs_uid:
                notice_text += "Cookies缺少 bbs_uid（例如 ltuid, stuid）"
            elif login_status.missing_login_ticket:
                notice_text += "Cookies缺少 login_ticket！"
            elif login_status.missing_cookie_token:
                notice_text += "Cookies缺少 cookie_token！"
            elif login_status.missing_stoken:
                notice_text += "Cookies缺少 stoken！"
            elif login_status.missing_stoken_v1:
                notice_text += "Cookies缺少 stoken_v1"
            elif login_status.missing_stoken_v2:
                notice_text += "Cookies缺少 stoken_v2"
            elif login_status.missing_mid:
                notice_text += "Cookies缺少 mid"
            else:
                notice_text += "未知错误！"
            notice_text += "[/] 如果部分步骤成功，你仍然可以尝试获取收货地址、兑换等功能"
            self.button.error.show()
            self.app.notice(notice_text)

        self.close_login()
        return login_status

    async def on_input_submitted(self, _: Input.Submitted) -> None:
        await self.login()

    async def _on_button_pressed(self, event: ControllableButton.Pressed) -> None:
        if event.button.id == "login":
            # 按下“登录”按钮时触发的事件

            await self.login()

        elif event.button.id in ["login_error", "login_success"]:
            # 按下“完成（成功）”或“返回（出错）”按钮时触发的事件

            if event.button.id == "login_success":
                [i.turn_off() for i in CaptchaLoginInformation.radio_tuple]
            self.button.login.show()
            self.button.error.hide()
            self.button.success.hide()
            self.before_login = True



from json import JSONDecodeError

import os
from httpx import Cookies
from loguru import logger
from pathlib import Path
from pydantic import BaseModel, Extra, ValidationError, BaseSettings, validator
from typing import List, Union, Optional, Tuple, Any, Dict, Set, Callable, TYPE_CHECKING, AbstractSet, \
    Mapping

from mys_goods_tool.data_model import BaseModelWithSetter, Good, Address, GameRecord, BaseModelWithUpdate

ROOT_PATH = Path("./")
"""程序所在目录"""

CONFIG_PATH = ROOT_PATH / "user_data.json"
"""用户数据文件默认路径"""

VERSION = "2.0.5"
"""程序当前版本"""

if TYPE_CHECKING:
    IntStr = Union[int, str]
    DictStrAny = Dict[str, Any]
    AbstractSetIntStr = AbstractSet[IntStr]
    MappingIntStrAny = Mapping[IntStr, Any]


class BBSCookies(BaseModelWithSetter, BaseModelWithUpdate):
    """
    米游社Cookies数据

    # 测试 is_correct() 方法

    >>> assert BBSCookies().is_correct() is False
    >>> assert BBSCookies(stuid="123", stoken="123", cookie_token="123").is_correct() is True

    # 测试 bbs_uid getter

    >>> bbs_cookies = BBSCookies()
    >>> assert not bbs_cookies.bbs_uid
    >>> assert BBSCookies(stuid="123").bbs_uid == "123"

    # 测试 bbs_uid setter

    >>> bbs_cookies.bbs_uid = "123"
    >>> assert bbs_cookies.bbs_uid == "123"

    # 检查构造函数内所用的 stoken setter

    >>> bbs_cookies = BBSCookies(stoken="abcd1234")
    >>> assert bbs_cookies.stoken_v1 and not bbs_cookies.stoken_v2
    >>> bbs_cookies = BBSCookies(stoken="v2_abcd1234==")
    >>> assert bbs_cookies.stoken_v2 and not bbs_cookies.stoken_v1
    >>> assert bbs_cookies.stoken == "v2_abcd1234=="

    # 检查 stoken setter

    >>> bbs_cookies = BBSCookies(stoken="abcd1234")
    >>> bbs_cookies.stoken = "v2_abcd1234=="
    >>> assert bbs_cookies.stoken_v2 == "v2_abcd1234=="
    >>> assert bbs_cookies.stoken_v1 == "abcd1234"

    # 检查 .dict 方法能否生成包含 stoken_2 类型的 stoken 的字典

    >>> bbs_cookies = BBSCookies()
    >>> bbs_cookies.stoken_v1 = "abcd1234"
    >>> bbs_cookies.stoken_v2 = "v2_abcd1234=="
    >>> assert bbs_cookies.dict(v2_stoken=True)["stoken"] == "v2_abcd1234=="

    # 检查是否有多余的字段

    >>> bbs_cookies = BBSCookies(stuid="123")
    >>> assert all(bbs_cookies.dict())
    >>> assert all(map(lambda x: x not in bbs_cookies, ["stoken_v1", "stoken_v2"]))

    # 测试 update 方法

    >>> bbs_cookies = BBSCookies(stuid="123")
    >>> assert bbs_cookies.update({"stuid": "456", "stoken": "abc"}) is bbs_cookies
    >>> assert bbs_cookies.stuid == "456"
    >>> assert bbs_cookies.stoken == "abc"

    >>> bbs_cookies = BBSCookies(stuid="123")
    >>> new_cookies = BBSCookies(stuid="456", stoken="abc")
    >>> assert bbs_cookies.update(new_cookies) is bbs_cookies
    >>> assert bbs_cookies.stuid == "456"
    >>> assert bbs_cookies.stoken == "abc"
    """
    stuid: Optional[str]
    """米游社UID"""
    ltuid: Optional[str]
    """米游社UID"""
    account_id: Optional[str]
    """米游社UID"""
    login_uid: Optional[str]
    """米游社UID"""

    stoken_v1: Optional[str]
    """保存stoken_v1，方便后续使用"""
    stoken_v2: Optional[str]
    """保存stoken_v2，方便后续使用"""

    cookie_token: Optional[str]
    login_ticket: Optional[str]
    ltoken: Optional[str]
    mid: Optional[str]

    def __init__(self, **data: Any):
        super().__init__(**data)
        stoken = data.get("stoken")
        if stoken:
            self.stoken = stoken

    def is_correct(self) -> bool:
        """判断是否为正确的Cookies"""
        if self.bbs_uid and self.stoken and self.cookie_token:
            return True
        else:
            return False

    @property
    def bbs_uid(self):
        """
        获取米游社UID
        """
        uid = None
        for value in [self.stuid, self.ltuid, self.account_id, self.login_uid]:
            if value:
                uid = value
                break
        return uid or None

    @bbs_uid.setter
    def bbs_uid(self, value: str):
        self.stuid = value
        self.ltuid = value
        self.account_id = value
        self.login_uid = value

    @property
    def stoken(self):
        """
        获取stoken
        :return: 优先返回 self.stoken_v1
        """
        if self.stoken_v1:
            return self.stoken_v1
        elif self.stoken_v2:
            return self.stoken_v2
        else:
            return None

    @stoken.setter
    def stoken(self, value):
        if value.startswith("v2_"):
            self.stoken_v2 = value
        else:
            self.stoken_v1 = value

    def update(self, cookies: Union[Dict[str, str], Cookies, "BBSCookies"]):
        """
        更新Cookies
        """
        if not isinstance(cookies, BBSCookies):
            self.stoken = cookies.get("stoken") or self.stoken
            self.bbs_uid = cookies.get("bbs_uid") or self.bbs_uid
            cookies.pop("stoken", None)
            cookies.pop("bbs_uid", None)
        return super().update(cookies)

    def dict(self, *,
             include: Optional[Union['AbstractSetIntStr', 'MappingIntStrAny']] = None,
             exclude: Optional[Union['AbstractSetIntStr', 'MappingIntStrAny']] = None,
             by_alias: bool = False,
             skip_defaults: Optional[bool] = None, exclude_unset: bool = False, exclude_defaults: bool = False,
             exclude_none: bool = False, v2_stoken: bool = False,
             cookie_type: bool = False) -> 'DictStrAny':
        """
        获取Cookies字典

        v2_stoken: stoken 字段是否使用 stoken_v2
        cookie_type: 是否返回符合Cookie类型的字典（没有自定义的stoken_v1、stoken_v2键）
        """
        # 保证 stuid, ltuid 等字段存在
        self.bbs_uid = self.bbs_uid
        cookies_dict = super().dict(include=include, exclude=exclude, by_alias=by_alias, skip_defaults=skip_defaults,
                                    exclude_unset=exclude_unset, exclude_defaults=exclude_defaults,
                                    exclude_none=exclude_none)
        if v2_stoken:
            cookies_dict["stoken"] = self.stoken_v2

        if cookie_type:
            # 去除自定义的 stoken_v1, stoken_v2 字段
            cookies_dict.pop("stoken_v1")
            cookies_dict.pop("stoken_v2")

            # 去除空的字段
            empty_key = set()
            for key, value in cookies_dict.items():
                if not value:
                    empty_key.add(key)
            [cookies_dict.pop(key) for key in empty_key]

        return cookies_dict


class UserAccount(BaseModelWithSetter, extra=Extra.ignore):
    """
    米游社账户数据

    >>> user_account = UserAccount(cookies=BBSCookies())
    >>> assert isinstance(user_account, UserAccount)
    >>> user_account.bbs_uid = "123"
    >>> assert user_account.bbs_uid == "123"
    """
    phone_number: Optional[str]
    """手机号"""
    cookies: BBSCookies
    """Cookies"""

    device_id_ios: str
    """iOS设备用 deviceID"""
    device_id_android: str
    """安卓设备用 deviceID"""

    def __init__(self, **data: Any):
        if not data.get("device_id_ios") or not data.get("device_id_android"):
            from mys_goods_tool.utils import generate_device_id
            if not data.get("device_id_ios"):
                data.setdefault("device_id_ios", generate_device_id())
            if not data.get("device_id_android"):
                data.setdefault("device_id_android", generate_device_id())
        super().__init__(**data)

    @property
    def bbs_uid(self):
        """
        获取米游社UID
        """
        return self.cookies.bbs_uid

    @bbs_uid.setter
    def bbs_uid(self, value: str):
        self.cookies.bbs_uid = value


class ExchangePlan(BaseModel):
    """
    兑换计划数据类
    """
    good: Good
    """商品"""
    address: Optional[Address]
    """地址ID"""
    account: UserAccount
    """米游社账号"""
    game_record: Optional[GameRecord]
    """商品对应的游戏的玩家账号"""

    def __hash__(self):
        return hash(
            (
                self.good.goods_id,
                self.good.time,
                self.address.id if self.address else None,
                self.account.bbs_uid,
                self.game_record.game_role_id if self.game_record else None
            )
        )


class ExchangeResult(BaseModel):
    """
    兑换结果数据类
    """
    result: bool
    """兑换结果"""
    return_data: dict
    """返回数据"""
    plan: ExchangePlan
    """兑换计划"""


class Preference(BaseSettings):
    """
    偏好设置
    """
    github_proxy: Optional[str] = "https://ghproxy.com/"
    """GitHub加速代理 最终会拼接在原GitHub链接前面"""
    enable_connection_test: bool = True
    """是否开启连接测试"""
    connection_test_interval: Optional[float] = 30
    """连接测试间隔（单位：秒）"""
    timeout: float = 10
    """网络请求超时时间（单位：秒）"""
    max_retry_times: Optional[int] = 3
    """最大网络请求重试次数"""
    retry_interval: float = 2
    """网络请求重试间隔（单位：秒）（除兑换请求外）"""
    enable_ntp_sync: Optional[bool] = True
    """是否开启NTP时间同步（将调整实际发出兑换请求的时间，而不是修改系统时间）"""
    ntp_server: Optional[str] = "ntp.aliyun.com"
    """NTP服务器地址"""
    timezone: Optional[str] = "Asia/Shanghai"
    """兑换时所用的时区"""
    geetest_statics_path: Optional[Path]
    """GEETEST行为验证 网站静态文件目录（默认读取本地包自带的静态文件）"""
    geetest_listen_address: Optional[Tuple[str, int]] = ("localhost", 0)
    """登录时使用的 GEETEST行为验证 WEB服务 本地监听地址"""
    exchange_thread_count: int = 2
    """兑换线程数"""
    exchange_latency: Tuple[float, float] = (0, 0.2)
    """兑换时间延迟随机范围（单位：秒）（防止因为发出请求的时间过于精准而被服务器认定为非人工操作）"""
    enable_log_output: bool = True
    """是否保存日志"""
    log_path: Optional[Path] = ROOT_PATH / "logs" / "mys_goods_tool.log"
    """日志保存路径"""

    @validator("log_path")
    def _(cls, v: Optional[Path]):
        absolute_path = v.absolute()
        if not os.path.exists(absolute_path) or not os.path.isfile(absolute_path):
            absolute_parent = absolute_path.parent
            try:
                os.makedirs(absolute_parent, exist_ok=True)
            except PermissionError:
                logger.warning(f"程序没有创建日志目录 {absolute_parent} 的权限")
        elif not os.access(absolute_path, os.W_OK):
            logger.warning(f"程序没有写入日志文件 {absolute_path} 的权限")
        return v

    class Config:
        env_prefix = "MYS_GOODS_TOOL_"  # 环境变量前缀


class SaltConfig(BaseSettings):
    """
    生成Headers - DS所用salt值
    """
    SALT_IOS: str = "ulInCDohgEs557j0VsPDYnQaaz6KJcv5"
    '''生成Headers iOS DS所需的salt'''
    SALT_ANDROID: str = "n0KjuIrKgLHh08LWSCYP0WXlVXaYvV64"
    '''生成Headers Android DS所需的salt'''
    SALT_DATA: str = "t0qEgfub6cvueAPgR5m9aQWWVciEer7v"
    '''Android 设备传入content生成 DS 所需的 salt'''
    SALT_PARAMS: str = "xV8v4Qu54lUKrEYFZkJhB8cuOh9Asafs"
    '''Android 设备传入url参数生成 DS 所需的 salt'''
    SALT_PROD: str = "JwYDpKvLj6MrMqqYU6jTKF17KNO2PXoS"

    class Config(Preference.Config):
        pass


# TODO 与用户数据文件中的配置有差异时询问是否使用更新
class DeviceConfig(BaseSettings):
    """
    设备信息
    DS算法与设备信息有关联，非必要请勿修改
    """
    USER_AGENT_MOBILE: str = "Mozilla/5.0 (iPhone; CPU iPhone OS 15_4 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) miHoYoBBS/2.42.1"
    '''移动端 User-Agent(Mozilla UA)'''
    USER_AGENT_PC: str = "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/16.0 Safari/605.1.15"
    '''桌面端 User-Agent(Mozilla UA)'''
    USER_AGENT_OTHER: str = "Hyperion/275 CFNetwork/1402.0.8 Darwin/22.2.0"
    '''获取用户 ActionTicket 时Headers所用的 User-Agent'''
    USER_AGENT_ANDROID: str = "Mozilla/5.0 (Linux; Android 11; MI 8 SE Build/RQ3A.211001.001; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/104.0.5112.97 Mobile Safari/537.36 miHoYoBBS/2.36.1"
    '''安卓端 User-Agent(Mozilla UA)'''
    USER_AGENT_ANDROID_OTHER: str = "okhttp/4.9.3"
    '''安卓端 User-Agent(专用于米游币任务等)'''
    USER_AGENT_WIDGET: str = "WidgetExtension/231 CFNetwork/1390 Darwin/22.0.0"
    '''iOS 小组件 User-Agent(原神实时便笺)'''

    X_RPC_DEVICE_MODEL_MOBILE: str = "iPhone10,2"
    '''移动端 x-rpc-device_model'''
    X_RPC_DEVICE_MODEL_PC: str = "OS X 10.15.7"
    '''桌面端 x-rpc-device_model'''
    X_RPC_DEVICE_MODEL_ANDROID: str = "MI 8 SE"
    '''安卓端 x-rpc-device_model'''

    X_RPC_DEVICE_NAME_MOBILE: str = "iPhone"
    '''移动端 x-rpc-device_name'''
    X_RPC_DEVICE_NAME_PC: str = "Microsoft Edge 103.0.1264.62"
    '''桌面端 x-rpc-device_name'''
    X_RPC_DEVICE_NAME_ANDROID: str = "Xiaomi MI 8 SE"
    '''安卓端 x-rpc-device_name'''

    X_RPC_SYS_VERSION: str = "15.4"
    '''Headers所用的 x-rpc-sys_version'''
    X_RPC_SYS_VERSION_ANDROID: str = "11"
    '''安卓端 x-rpc-sys_version'''

    X_RPC_CHANNEL: str = "appstore"
    '''Headers所用的 x-rpc-channel'''
    X_RPC_CHANNEL_ANDROID: str = "miyousheluodi"
    '''安卓端 x-rpc-channel'''

    X_RPC_APP_VERSION: str = "2.52.1"
    '''Headers所用的 x-rpc-app_version'''
    X_RPC_PLATFORM: str = "ios"
    '''Headers所用的 x-rpc-platform'''
    UA: str = "\".Not/A)Brand\";v=\"99\", \"Microsoft Edge\";v=\"103\", \"Chromium\";v=\"103\""
    '''Headers所用的 sec-ch-ua'''
    UA_PLATFORM: str = "\"macOS\""
    '''Headers所用的 sec-ch-ua-platform'''

    class Config(Preference.Config):
        pass


class UserData(BaseModel):
    """
    用户数据类
    """
    version: str = VERSION
    """创建用户数据文件的程序版本号"""
    exchange_plans: Union[Set[ExchangePlan], List[ExchangePlan]] = set()
    """兑换计划列表"""
    preference: Preference = Preference()
    """偏好设置"""
    salt_config: SaltConfig = SaltConfig()
    """生成Headers - DS所用salt值"""
    device_config: DeviceConfig = DeviceConfig()
    """设备信息"""
    accounts: Dict[str, UserAccount] = {}
    """储存一些已绑定的账号数据"""

    def __init__(self, **data: Any):
        super().__init__(**data)
        exchange_plans = self.exchange_plans
        self.exchange_plans = set()
        for plan in exchange_plans:
            plan = ExchangePlan.parse_obj(plan)
            self.exchange_plans.add(plan)

    def save(self):
        """
        保存用户数据文件
        """
        return write_config_file(self)

    def json(
            self,
            *,
            include: Optional[Union['AbstractSetIntStr', 'MappingIntStrAny']] = None,
            exclude: Optional[Union['AbstractSetIntStr', 'MappingIntStrAny']] = None,
            by_alias: bool = False,
            skip_defaults: Optional[bool] = None,
            exclude_unset: bool = False,
            exclude_defaults: bool = False,
            exclude_none: bool = False,
            encoder: Optional[Callable[[Any], Any]] = None,
            models_as_dict: bool = True,
            **dumps_kwargs: Any,
    ) -> str:
        """
        重写 BaseModel.json() 方法，使其支持对 Set 类型的数据进行序列化
        """
        set_plans = self.exchange_plans
        self.exchange_plans = list(set_plans)
        json_data = super().json(
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            skip_defaults=skip_defaults,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            encoder=encoder,
            models_as_dict=models_as_dict,
            **dumps_kwargs,
        )
        self.exchange_plans = set_plans
        return json_data


def write_config_file(conf: UserData = UserData()):
    """
    写入用户数据文件

    :param conf: 配置对象
    """
    try:
        str_data = conf.json(indent=4)
    except (AttributeError, TypeError, ValueError):
        logger.exception("数据对象序列化失败，可能是数据类型错误")
        return False
    with open(CONFIG_PATH, "w", encoding="utf-8") as f:
        f.write(str_data)
    return True


def load_config():
    """
    加载用户数据文件
    """
    if os.path.exists(CONFIG_PATH) and os.path.isfile(CONFIG_PATH):
        try:
            return UserData.parse_file(CONFIG_PATH)
        except (ValidationError, JSONDecodeError):
            logger.exception(f"读取用户数据文件失败，请检查用户数据文件 {CONFIG_PATH} 格式是否正确")
            exit(1)
        except:
            logger.exception(f"读取用户数据文件失败，请检查用户数据文件 {CONFIG_PATH} 是否存在且程序有权限读取和写入")
            exit(1)
    else:
        user_data = UserData()
        try:
            write_config_file(user_data)
        except PermissionError:
            logger.exception(f"创建用户数据文件失败，请检查程序是否有权限读取和写入 {CONFIG_PATH}")
            exit(1)
        # logger.info(f"用户数据文件 {CONFIG_PATH} 不存在，已创建默认用户数据文件。")
        # 由于会输出到标准输出流，影响TUI观感，因此暂时取消

        return user_data


config = load_config()
"""程序配置对象"""
